@@ -13,7 +13,7 @@ typedef struct sav_file_header_record_s {
     char     prod_name[60];
     int32_t  layout_code;
     int32_t  nominal_case_size;
-    int32_t  compressed;
+    int32_t  compression;
     int32_t  weight_index;
     int32_t  ncases;
     double   bias; /* TODO is this portable? */
@@ -64,16 +64,12 @@ typedef struct sav_dictionary_termination_record_s {
 #pragma pack(pop)
 
 typedef struct sav_ctx_s {
-    readstat_error_handler          error_handler;
-    readstat_progress_handler       progress_handler;
-    readstat_note_handler           note_handler;
-    readstat_value_handler          value_handler;
-    readstat_value_label_handler    value_label_handler;
-    size_t                          file_size;
-    readstat_io_t                  *io;
-    void                           *user_ctx;
+    readstat_callbacks_t  handle;
+    size_t                file_size;
+    readstat_io_t        *io;
+    void                 *user_ctx;
 
-    spss_varinfo_t       *varinfo;
+    spss_varinfo_t      **varinfo;
     size_t                varinfo_capacity;
     readstat_variable_t **variables;
 
@@ -81,8 +77,8 @@ typedef struct sav_ctx_s {
     const char    *output_encoding;
     char           file_label[4*64+1];
     time_t         timestamp;
-    int32_t       *variable_display_values;
-    int            variable_display_values_count;
+    uint32_t      *variable_display_values;
+    size_t         variable_display_values_count;
     iconv_t        converter;
     int            var_index;
     int            var_offset;
@@ -103,7 +99,11 @@ typedef struct sav_ctx_s {
     uint64_t       lowest_double;
     uint64_t       highest_double;
 
-    unsigned int   data_is_compressed:1;
+    double         bias;
+    int            format_version;
+
+    readstat_compress_t  compression;
+    readstat_endian_t   endianness;
     unsigned int   bswap:1;
 } sav_ctx_t;
 
@@ -116,6 +116,7 @@ typedef struct sav_ctx_s {
 
 #define SAV_RECORD_SUBTYPE_INTEGER_INFO       3
 #define SAV_RECORD_SUBTYPE_FP_INFO            4
+#define SAV_RECORD_SUBTYPE_PRODUCT_INFO      10
 #define SAV_RECORD_SUBTYPE_VAR_DISPLAY       11
 #define SAV_RECORD_SUBTYPE_LONG_VAR_NAME     13
 #define SAV_RECORD_SUBTYPE_VERY_LONG_STR     14
@@ -132,11 +133,7 @@ typedef struct sav_ctx_s {
 #define SAV_ENDIANNESS_BIG               1
 #define SAV_ENDIANNESS_LITTLE            2
 
-#define SAV_CHARSET_EBCDIC                1
-#define SAV_CHARSET_7_BIT_ASCII           2
-#define SAV_CHARSET_8_BIT_ASCII           3
-#define SAV_CHARSET_DEC_KANJI             4
-#define SAV_CHARSET_UTF8              65001
+#define SAV_EIGHT_SPACES              "        "
 
 sav_ctx_t *sav_ctx_init(sav_file_header_record_t *header, readstat_io_t *io);
 void sav_ctx_free(sav_ctx_t *ctx);
