@@ -15,9 +15,14 @@
 #include "../readstat_malloc.h"
 
 #include "readstat_sav.h"
+#include "readstat_sav_compress.h"
 #include "readstat_sav_parse.h"
 #include "readstat_sav_parse_timestamp.h"
 
+#if HAVE_ZLIB
+#include "readstat_zsav_read.h"
+#endif
+
 #define DATA_BUFFER_SIZE    65536
 
 /* Others defined in table below */
@@ -108,15 +113,18 @@ static readstat_charset_entry_t _charset_table[] = {
 #define SAV_LABEL_NAME_PREFIX         "labels"
 
 typedef struct value_label_s {
-    char          value[8];
-    unsigned char label_len;
-    char          label[256*4+1];
+    char             raw_value[8];
+    char             utf8_string_value[8*4+1];
+    readstat_value_t final_value;
+    char            *label;
 } value_label_t;
 
 static readstat_error_t sav_update_progress(sav_ctx_t *ctx);
 static readstat_error_t sav_read_data(sav_ctx_t *ctx);
-static readstat_error_t sav_read_compressed_data(sav_ctx_t *ctx);
-static readstat_error_t sav_read_uncompressed_data(sav_ctx_t *ctx);
+static readstat_error_t sav_read_compressed_data(sav_ctx_t *ctx,
+        readstat_error_t (*row_handler)(unsigned char *, size_t, sav_ctx_t *));
+static readstat_error_t sav_read_uncompressed_data(sav_ctx_t *ctx,
+        readstat_error_t (*row_handler)(unsigned char *, size_t, sav_ctx_t *));
 
 static readstat_error_t sav_skip_variable_record(sav_ctx_t *ctx);
 static readstat_error_t sav_read_variable_record(sav_ctx_t *ctx);
@@ -133,7 +141,7 @@ static readstat_error_t sav_parse_machine_floating_point_record(const void *data
 static readstat_error_t sav_store_variable_display_parameter_record(const void *data, size_t size, size_t count, sav_ctx_t *ctx);
 static readstat_error_t sav_parse_variable_display_parameter_record(sav_ctx_t *ctx);
 static readstat_error_t sav_parse_machine_integer_info_record(const void *data, size_t data_len, sav_ctx_t *ctx);
-static readstat_error_t sav_parse_long_value_labels_record(const void *data, size_t data_len, sav_ctx_t *ctx);
+static readstat_error_t sav_parse_long_value_labels_record(const void *data, size_t size, size_t count, sav_ctx_t *ctx);
 
 static void sav_tag_missing_double(readstat_value_t *value, sav_ctx_t *ctx) {
     double fp_value = value->v.double_value;
@@ -151,7 +159,7 @@ static void sav_tag_missing_double(readstat_value_t *value, sav_ctx_t *ctx) {
 
 static readstat_error_t sav_update_progress(sav_ctx_t *ctx) {
     readstat_io_t *io = ctx->io;
-    return io->update(ctx->file_size, ctx->progress_handler, ctx->user_ctx, io->io_ctx);
+    return io->update(ctx->file_size, ctx->handle.progress, ctx->user_ctx, io->io_ctx);
 }
 
 static readstat_error_t sav_skip_variable_record(sav_ctx_t *ctx) {
@@ -198,6 +206,9 @@ static readstat_error_t sav_read_variable_label(spss_varinfo_t *info, sav_ctx_t
     }
     label_len = ctx->bswap ? byteswap4(label_len) : label_len;
 
+    if (label_len == 0)
+        goto cleanup;
+
     label_capacity = (label_len + 3) / 4 * 4;
     if ((label_buf = readstat_malloc(label_capacity)) == NULL) {
         retval = READSTAT_ERROR_MALLOC;
@@ -212,8 +223,6 @@ static readstat_error_t sav_read_variable_label(spss_varinfo_t *info, sav_ctx_t
 
     if (io->read(label_buf, label_capacity, io->io_ctx) < label_capacity) {
         retval = READSTAT_ERROR_READ;
-        free(info->label);
-        info->label = NULL;
         goto cleanup;
     }
 
@@ -225,6 +234,13 @@ cleanup:
     if (label_buf)
         free(label_buf);
 
+    if (retval != READSTAT_OK) {
+        if (info->label) {
+            free(info->label);
+            info->label = NULL;
+        }
+    }
+
     return retval;
 }
 
@@ -270,10 +286,11 @@ cleanup:
 
 static readstat_error_t sav_read_variable_record(sav_ctx_t *ctx) {
     readstat_io_t *io = ctx->io;
-    sav_variable_record_t variable;
+    sav_variable_record_t variable = { 0 };
+    spss_varinfo_t *info = NULL;
     readstat_error_t retval = READSTAT_OK;
     if (ctx->var_index == ctx->varinfo_capacity) {
-        if ((ctx->varinfo = readstat_realloc(ctx->varinfo, (ctx->varinfo_capacity *= 2) * sizeof(spss_varinfo_t))) == NULL) {
+        if ((ctx->varinfo = readstat_realloc(ctx->varinfo, (ctx->varinfo_capacity *= 2) * sizeof(spss_varinfo_t *))) == NULL) {
             retval = READSTAT_ERROR_MALLOC;
             goto cleanup;
         }
@@ -291,12 +308,14 @@ static readstat_error_t sav_read_variable_record(sav_ctx_t *ctx) {
             return READSTAT_ERROR_PARSE;
         }
         ctx->var_offset++;
-        spss_varinfo_t *prev = &ctx->varinfo[ctx->var_index-1];
-        prev->width++;
+        ctx->varinfo[ctx->var_index-1]->width++;
         return 0;
     }
-    spss_varinfo_t *info = &ctx->varinfo[ctx->var_index];
-    memset(info, 0, sizeof(spss_varinfo_t));
+
+    if ((info = readstat_calloc(1, sizeof(spss_varinfo_t))) == NULL) {
+        retval = READSTAT_ERROR_MALLOC;
+        goto cleanup;
+    }
     info->width = 1;
     info->n_segments = 1;
     info->index = ctx->var_index;
@@ -340,11 +359,17 @@ static readstat_error_t sav_read_variable_record(sav_ctx_t *ctx) {
         }
     }
     
+    ctx->varinfo[ctx->var_index] = info;
+
     ctx->var_index++;
     ctx->var_offset++;
     
 cleanup:
-    
+    if (retval != READSTAT_OK) {
+        if (info)
+            free(info);
+    }
+
     return retval;
 }
 
@@ -362,13 +387,18 @@ static readstat_error_t sav_skip_value_label_record(sav_ctx_t *ctx) {
         label_count = byteswap4(label_count);
     int i;
     for (i=0; i<label_count; i++) {
-        value_label_t vlabel;
-        if (io->read(&vlabel, 9, io->io_ctx) < 9) {
+        unsigned char unpadded_len = 0;
+        size_t padded_len = 0;
+        if (io->seek(8, READSTAT_SEEK_CUR, io->io_ctx) == -1) {
+            retval = READSTAT_ERROR_SEEK;
+            goto cleanup;
+        }
+        if (io->read(&unpadded_len, 1, io->io_ctx) < 1) {
             retval = READSTAT_ERROR_READ;
             goto cleanup;
         }
-        size_t label_len = (vlabel.label_len + 8) / 8 * 8 - 1;
-        if (io->seek(label_len, READSTAT_SEEK_CUR, io->io_ctx) == -1) {
+        padded_len = (unpadded_len + 8) / 8 * 8 - 1;
+        if (io->seek(padded_len, READSTAT_SEEK_CUR, io->io_ctx) == -1) {
             retval = READSTAT_ERROR_SEEK;
             goto cleanup;
         }
@@ -411,24 +441,7 @@ static readstat_error_t sav_submit_value_labels(value_label_t *value_labels, int
 
     for (i=0; i<label_count; i++) {
         value_label_t *vlabel = &value_labels[i];
-        readstat_value_t value = { .type = value_type };
-        double val_d = 0.0;
-        char unpadded_val[8*4+1];
-        if (value_type == READSTAT_TYPE_DOUBLE) {
-            memcpy(&val_d, vlabel->value, 8);
-            if (ctx->bswap)
-                val_d = byteswap_double(val_d);
-
-            value.v.double_value = val_d;
-            sav_tag_missing_double(&value, ctx);
-        } else {
-            retval = readstat_convert(unpadded_val, sizeof(unpadded_val), vlabel->value, 8, ctx->converter);
-            if (retval != READSTAT_OK)
-                break;
-
-            value.v.string_value = unpadded_val;
-        }
-        if (ctx->value_label_handler(label_name_buf, value, vlabel->label, ctx->user_ctx) != READSTAT_HANDLER_OK) {
+        if (ctx->handle.value_label(label_name_buf, vlabel->final_value, vlabel->label, ctx->user_ctx) != READSTAT_HANDLER_OK) {
             retval = READSTAT_ERROR_USER_ABORT;
             goto cleanup;
         }
@@ -455,7 +468,7 @@ static readstat_error_t sav_read_value_label_record(sav_ctx_t *ctx) {
     if (ctx->bswap)
         label_count = byteswap4(label_count);
     
-    if (label_count && (value_labels = readstat_malloc(label_count * sizeof(value_label_t))) == NULL) {
+    if (label_count && (value_labels = readstat_calloc(label_count, sizeof(value_label_t))) == NULL) {
         retval = READSTAT_ERROR_MALLOC;
         goto cleanup;
     }
@@ -463,16 +476,31 @@ static readstat_error_t sav_read_value_label_record(sav_ctx_t *ctx) {
     int i;
     for (i=0; i<label_count; i++) {
         value_label_t *vlabel = &value_labels[i];
-        if (io->read(vlabel, 9, io->io_ctx) < 9) {
+        unsigned char unpadded_label_len = 0;
+        size_t padded_label_len = 0, utf8_label_len = 0;
+
+        if (io->read(vlabel->raw_value, 8, io->io_ctx) < 8) {
             retval = READSTAT_ERROR_READ;
             goto cleanup;
         }
-        size_t label_len = (vlabel->label_len + 8) / 8 * 8 - 1;
-        if (io->read(label_buf, label_len, io->io_ctx) < label_len) {
+        if (io->read(&unpadded_label_len, 1, io->io_ctx) < 1) {
             retval = READSTAT_ERROR_READ;
             goto cleanup;
         }
-        retval = readstat_convert(vlabel->label, sizeof(vlabel->label), label_buf, label_len, ctx->converter);
+
+        padded_label_len = (unpadded_label_len + 8) / 8 * 8 - 1;
+        if (io->read(label_buf, padded_label_len, io->io_ctx) < padded_label_len) {
+            retval = READSTAT_ERROR_READ;
+            goto cleanup;
+        }
+
+        utf8_label_len = padded_label_len*4+1;
+        if ((vlabel->label = readstat_malloc(utf8_label_len)) == NULL) {
+            retval = READSTAT_ERROR_MALLOC;
+            goto cleanup;
+        }
+
+        retval = readstat_convert(vlabel->label, utf8_label_len, label_buf, padded_label_len, ctx->converter);
         if (retval != READSTAT_OK)
             goto cleanup;
     }
@@ -509,23 +537,50 @@ static readstat_error_t sav_read_value_label_record(sav_ctx_t *ctx) {
             var_offset = byteswap4(var_offset);
 
         var_offset--; // Why subtract 1????
-        spss_varinfo_t *var = bsearch(&var_offset, ctx->varinfo, ctx->var_index, sizeof(spss_varinfo_t),
+        spss_varinfo_t **var = bsearch(&var_offset, ctx->varinfo, ctx->var_index, sizeof(spss_varinfo_t *),
                 &spss_varinfo_compare);
         if (var) {
-            var->labels_index = ctx->value_labels_count;
-
-            value_type = var->type;
+            (*var)->labels_index = ctx->value_labels_count;
+            value_type = (*var)->type;
         }
     }
-    if (ctx->value_label_handler) {
+
+    for (i=0; i<label_count; i++) {
+        value_label_t *vlabel = &value_labels[i];
+        double val_d = 0.0;
+        vlabel->final_value.type = value_type;
+        if (value_type == READSTAT_TYPE_DOUBLE) {
+            memcpy(&val_d, vlabel->raw_value, 8);
+            if (ctx->bswap)
+                val_d = byteswap_double(val_d);
+
+            vlabel->final_value.v.double_value = val_d;
+            sav_tag_missing_double(&vlabel->final_value, ctx);
+        } else {
+            retval = readstat_convert(vlabel->utf8_string_value, sizeof(vlabel->utf8_string_value),
+                    vlabel->raw_value, 8, ctx->converter);
+            if (retval != READSTAT_OK)
+                break;
+
+            vlabel->final_value.v.string_value = vlabel->utf8_string_value;
+        }
+    }
+
+    if (ctx->handle.value_label) {
         sav_submit_value_labels(value_labels, label_count, value_type, ctx);
     }
     ctx->value_labels_count++;
 cleanup:
     if (vars)
         free(vars);
-    if (value_labels)
+    if (value_labels) {
+        for (i=0; i<label_count; i++) {
+            value_label_t *vlabel = &value_labels[i];
+            if (vlabel->label)
+                free(vlabel->label);
+        }
         free(value_labels);
+    }
     
     return retval;
 }
@@ -550,7 +605,7 @@ cleanup:
 }
 
 static readstat_error_t sav_read_document_record(sav_ctx_t *ctx) {
-    if (!ctx->note_handler)
+    if (!ctx->handle.note)
         return sav_skip_document_record(ctx);
 
     uint32_t n_lines;
@@ -577,7 +632,7 @@ static readstat_error_t sav_read_document_record(sav_ctx_t *ctx) {
         if (retval != READSTAT_OK)
             goto cleanup;
 
-        if (ctx->note_handler(i, utf8_buffer, ctx->user_ctx) != READSTAT_HANDLER_OK) {
+        if (ctx->handle.note(i, utf8_buffer, ctx->user_ctx) != READSTAT_HANDLER_OK) {
             retval = READSTAT_ERROR_USER_ABORT;
             goto cleanup;
         }
@@ -597,45 +652,6 @@ static readstat_error_t sav_read_dictionary_termination_record(sav_ctx_t *ctx) {
     return retval;
 }
 
-static readstat_error_t sav_read_data(sav_ctx_t *ctx) {
-    readstat_error_t retval = READSTAT_OK;
-    int longest_string = 256;
-    int i;
-
-    for (i=0; i<ctx->var_count; i++) {
-        spss_varinfo_t *info = &ctx->varinfo[i];
-        if (info->string_length > longest_string) {
-            longest_string = info->string_length;
-        }
-    }
-
-    ctx->raw_string_len = longest_string + sizeof(SAV_EIGHT_SPACES)-2;
-    ctx->raw_string = readstat_malloc(ctx->raw_string_len);
-
-    ctx->utf8_string_len = 4*longest_string+1 + sizeof(SAV_EIGHT_SPACES)-2;
-    ctx->utf8_string = readstat_malloc(ctx->utf8_string_len);
-
-    if (ctx->raw_string == NULL || ctx->utf8_string == NULL) {
-        retval = READSTAT_ERROR_MALLOC;
-        goto done;
-    }
-
-    if (ctx->data_is_compressed) {
-        retval = sav_read_compressed_data(ctx);
-    } else {
-        retval = sav_read_uncompressed_data(ctx);
-    }
-    if (retval != READSTAT_OK)
-        goto done;
-
-    if (ctx->record_count != -1 && ctx->current_row != ctx->row_limit) {
-        retval = READSTAT_ERROR_ROW_COUNT_MISMATCH;
-    }
-
-done:
-    return retval;
-}
-
 static readstat_error_t sav_process_row(unsigned char *buffer, size_t buffer_len, sav_ctx_t *ctx) {
     readstat_error_t retval = READSTAT_OK;
     double fp_value;
@@ -646,8 +662,8 @@ static readstat_error_t sav_process_row(unsigned char *buffer, size_t buffer_len
     int var_index = 0, col = 0;
 
     while (data_offset < buffer_len && col < ctx->var_index) {
-        spss_varinfo_t *col_info = &ctx->varinfo[col];
-        spss_varinfo_t *var_info = &ctx->varinfo[var_index];
+        spss_varinfo_t *col_info = ctx->varinfo[col];
+        spss_varinfo_t *var_info = ctx->varinfo[var_index];
         readstat_value_t value = { .type = var_info->type };
         if (offset > 31) {
             retval = READSTAT_ERROR_PARSE;
@@ -672,7 +688,7 @@ static readstat_error_t sav_process_row(unsigned char *buffer, size_t buffer_len
                     if (retval != READSTAT_OK)
                         goto done;
                     value.v.string_value = ctx->utf8_string;
-                    if (ctx->value_handler(ctx->current_row, ctx->variables[var_info->index],
+                    if (ctx->handle.value(ctx->current_row, ctx->variables[var_info->index],
                                 value, ctx->user_ctx) != READSTAT_HANDLER_OK) {
                         retval = READSTAT_ERROR_USER_ABORT;
                         goto done;
@@ -690,7 +706,7 @@ static readstat_error_t sav_process_row(unsigned char *buffer, size_t buffer_len
                 }
                 value.v.double_value = fp_value;
                 sav_tag_missing_double(&value, ctx);
-                if (ctx->value_handler(ctx->current_row, ctx->variables[var_info->index],
+                if (ctx->handle.value(ctx->current_row, ctx->variables[var_info->index],
                             value, ctx->user_ctx) != READSTAT_HANDLER_OK) {
                     retval = READSTAT_ERROR_USER_ABORT;
                     goto done;
@@ -706,7 +722,54 @@ done:
     return retval;
 }
 
-static readstat_error_t sav_read_uncompressed_data(sav_ctx_t *ctx) {
+static readstat_error_t sav_read_data(sav_ctx_t *ctx) {
+    readstat_error_t retval = READSTAT_OK;
+    size_t longest_string = 256;
+    int i;
+
+    for (i=0; i<ctx->var_index;) {
+        spss_varinfo_t *info = ctx->varinfo[i];
+        if (info->string_length > longest_string) {
+            longest_string = info->string_length;
+        }
+        i += info->n_segments;
+    }
+
+    ctx->raw_string_len = longest_string + sizeof(SAV_EIGHT_SPACES)-2;
+    ctx->raw_string = readstat_malloc(ctx->raw_string_len);
+
+    ctx->utf8_string_len = 4*longest_string+1 + sizeof(SAV_EIGHT_SPACES)-2;
+    ctx->utf8_string = readstat_malloc(ctx->utf8_string_len);
+
+    if (ctx->raw_string == NULL || ctx->utf8_string == NULL) {
+        retval = READSTAT_ERROR_MALLOC;
+        goto done;
+    }
+
+    if (ctx->compression == READSTAT_COMPRESS_ROWS) {
+        retval = sav_read_compressed_data(ctx, &sav_process_row);
+    } else if (ctx->compression == READSTAT_COMPRESS_BINARY) {
+#if HAVE_ZLIB
+        retval = zsav_read_compressed_data(ctx, &sav_process_row);
+#else
+        retval = READSTAT_ERROR_UNSUPPORTED_COMPRESSION;
+#endif
+    } else {
+        retval = sav_read_uncompressed_data(ctx, &sav_process_row);
+    }
+    if (retval != READSTAT_OK)
+        goto done;
+
+    if (ctx->record_count != -1 && ctx->current_row != ctx->row_limit) {
+        retval = READSTAT_ERROR_ROW_COUNT_MISMATCH;
+    }
+
+done:
+    return retval;
+}
+
+static readstat_error_t sav_read_uncompressed_data(sav_ctx_t *ctx,
+        readstat_error_t (*row_handler)(unsigned char *, size_t, sav_ctx_t *)) {
     readstat_error_t retval = READSTAT_OK;
     readstat_io_t *io = ctx->io;
     unsigned char *buffer = NULL;
@@ -723,7 +786,7 @@ static readstat_error_t sav_read_uncompressed_data(sav_ctx_t *ctx) {
         if ((bytes_read = io->read(buffer, buffer_len, io->io_ctx)) != buffer_len)
             goto done;
 
-        retval = sav_process_row(buffer, buffer_len, ctx);
+        retval = row_handler(buffer, buffer_len, ctx);
         if (retval != READSTAT_OK)
             goto done;
     }
@@ -734,13 +797,10 @@ done:
     return retval;
 }
 
-static readstat_error_t sav_read_compressed_data(sav_ctx_t *ctx) {
+static readstat_error_t sav_read_compressed_data(sav_ctx_t *ctx,
+        readstat_error_t (*row_handler)(unsigned char *, size_t, sav_ctx_t *)) {
     readstat_error_t retval = READSTAT_OK;
     readstat_io_t *io = ctx->io;
-    unsigned char chunk[8];
-    int i;
-    double fp_value;
-    uint64_t missing_value = ctx->missing_double;
     readstat_off_t data_offset = 0;
     unsigned char buffer[DATA_BUFFER_SIZE];
     int buffer_used = 0;
@@ -749,8 +809,10 @@ static readstat_error_t sav_read_compressed_data(sav_ctx_t *ctx) {
     readstat_off_t uncompressed_offset = 0;
     unsigned char *uncompressed_row = NULL;
 
-    int bswap = ctx->bswap;
-    ctx->bswap = 0;
+    struct sav_row_stream_s state = { 
+        .missing_value = ctx->missing_double,
+        .bias = ctx->bias,
+        .bswap = ctx->bswap };
 
     if (uncompressed_row_len && (uncompressed_row = readstat_malloc(uncompressed_row_len)) == NULL) {
         retval = READSTAT_ERROR_MALLOC;
@@ -758,70 +820,42 @@ static readstat_error_t sav_read_compressed_data(sav_ctx_t *ctx) {
     }
 
     while (1) {
-        if (data_offset >= buffer_used) {
-            retval = sav_update_progress(ctx);
-            if (retval != READSTAT_OK)
-                goto done;
+        buffer_used = io->read(buffer, sizeof(buffer), io->io_ctx);
+        if (buffer_used == -1 || buffer_used == 0 || (buffer_used % 8) != 0)
+            goto done;
 
-            if ((buffer_used = io->read(buffer, sizeof(buffer), io->io_ctx)) == -1 ||
-                buffer_used == 0 || (buffer_used % 8) != 0)
-                goto done;
+        state.status = SAV_ROW_STREAM_HAVE_DATA;
+        data_offset = 0;
 
-            data_offset = 0;
-        }
+        while (state.status != SAV_ROW_STREAM_NEED_DATA) {
+            state.next_in = &buffer[data_offset];
+            state.avail_in = buffer_used - data_offset;
 
-        memcpy(chunk, &buffer[data_offset], 8);
-        data_offset += 8;
-        
-        for (i=0; i<8; i++) {
-            switch (chunk[i]) {
-                case 0:
-                    break;
-                case 252:
-                    goto done;
-                case 253:
-                    if (data_offset >= buffer_used) {
-                        if ((buffer_used = io->read(buffer, sizeof(buffer), io->io_ctx)) == -1 ||
-                            buffer_used == 0 || (buffer_used % 8) != 0)
-                            goto done;
+            state.next_out = &uncompressed_row[uncompressed_offset];
+            state.avail_out = uncompressed_row_len - uncompressed_offset;
 
-                        data_offset = 0;
-                    }
-                    memcpy(&uncompressed_row[uncompressed_offset], &buffer[data_offset], 8);
-                    uncompressed_offset += 8;
-                    data_offset += 8;
-                    break;
-                case 254:
-                    memcpy(&uncompressed_row[uncompressed_offset], SAV_EIGHT_SPACES, 8);
-                    uncompressed_offset += 8;
-                    break;
-                case 255:
-                    memcpy(&uncompressed_row[uncompressed_offset], &missing_value, sizeof(uint64_t));
-                    uncompressed_offset += 8;
-                    break;
-                default:
-                    fp_value = chunk[i] - 100.0;
-                    memcpy(&uncompressed_row[uncompressed_offset], &fp_value, sizeof(double));
-                    uncompressed_offset += 8;
-                    break;
-            }
-            if (uncompressed_offset == uncompressed_row_len) {
-                retval = sav_process_row(uncompressed_row, uncompressed_row_len, ctx);
+            sav_decompress_row(&state);
+
+            uncompressed_offset = uncompressed_row_len - state.avail_out;
+            data_offset = buffer_used - state.avail_in;
+
+            if (state.status == SAV_ROW_STREAM_FINISHED_ROW) {
+                retval = row_handler(uncompressed_row, uncompressed_row_len, ctx);
                 if (retval != READSTAT_OK)
                     goto done;
 
                 uncompressed_offset = 0;
             }
-            if (ctx->current_row == ctx->row_limit)
+
+            if (state.status == SAV_ROW_STREAM_FINISHED_ALL || ctx->current_row == ctx->row_limit)
                 goto done;
         }
     }
+
 done:
     if (uncompressed_row)
         free(uncompressed_row);
 
-    ctx->bswap = bswap;
-
     return retval;
 }
 
@@ -847,10 +881,10 @@ static readstat_error_t sav_parse_machine_integer_info_record(const void *data,
             }
         }
         if (src_charset == NULL) {
-            if (ctx->error_handler) {
+            if (ctx->handle.error) {
                 char error_buf[1024];
                 snprintf(error_buf, sizeof(error_buf), "Unsupported character set: %d\n", record.character_code);
-                ctx->error_handler(error_buf, ctx->user_ctx);
+                ctx->handle.error(error_buf, ctx->user_ctx);
             }
             return READSTAT_ERROR_UNSUPPORTED_CHARSET;
         }
@@ -861,6 +895,9 @@ static readstat_error_t sav_parse_machine_integer_info_record(const void *data,
         if (converter == (iconv_t)-1) {
             return READSTAT_ERROR_UNSUPPORTED_CHARSET;
         }
+        if (ctx->converter) {
+            iconv_close(ctx->converter);
+        }
         ctx->converter = converter;
     }
     return READSTAT_OK;
@@ -912,7 +949,7 @@ static readstat_error_t sav_parse_variable_display_parameter_record(sav_ctx_t *c
     int has_display_width = ctx->var_count > 0 && (count / ctx->var_count == 3);
     int offset = 0;
     for (i=0; i<ctx->var_index;) {
-        spss_varinfo_t *info = &ctx->varinfo[i];
+        spss_varinfo_t *info = ctx->varinfo[i];
         info->measure = spss_measure_to_readstat_measure(ctx->variable_display_values[offset++]);
         if (has_display_width) {
             info->display_width = ctx->variable_display_values[offset++];
@@ -924,16 +961,18 @@ static readstat_error_t sav_parse_variable_display_parameter_record(sav_ctx_t *c
     return READSTAT_OK;
 }
 
-static readstat_error_t sav_parse_long_value_labels_record(const void *data, size_t data_len, sav_ctx_t *ctx) {
-    if (!ctx->value_label_handler)
+static readstat_error_t sav_parse_long_value_labels_record(const void *data, size_t size, size_t count, sav_ctx_t *ctx) {
+    if (!ctx->handle.value_label)
         return READSTAT_OK;
+    if (size != 1)
+        return READSTAT_ERROR_PARSE;
 
     readstat_error_t retval = READSTAT_OK;
     uint32_t label_name_len = 0;
     uint32_t label_count = 0;
     uint32_t i = 0;
     const char *data_ptr = data;
-    const char *data_end = data_ptr + data_len;
+    const char *data_end = data_ptr + count;
     char var_name_buf[256*4+1];
     char label_name_buf[256];
     char *value_buffer = NULL;
@@ -964,7 +1003,7 @@ static readstat_error_t sav_parse_long_value_labels_record(const void *data, siz
     data_ptr += label_name_len;
 
     for (i=0; i<ctx->var_index;) {
-        spss_varinfo_t *info = &ctx->varinfo[i];
+        spss_varinfo_t *info = ctx->varinfo[i];
         if (strcmp(var_name_buf, info->longname) == 0) {
             info->labels_index = ctx->value_labels_count++;
             snprintf(label_name_buf, sizeof(label_name_buf),
@@ -1057,7 +1096,7 @@ static readstat_error_t sav_parse_long_value_labels_record(const void *data, siz
         readstat_value_t value = { .type = READSTAT_TYPE_STRING };
         value.v.string_value = value_buffer;
 
-        if (ctx->value_label_handler(label_name_buf, value, label_buffer, ctx->user_ctx) != READSTAT_HANDLER_OK) {
+        if (ctx->handle.value_label(label_name_buf, value, label_buffer, ctx->user_ctx) != READSTAT_HANDLER_OK) {
             retval = READSTAT_ERROR_USER_ABORT;
             goto cleanup;
         }
@@ -1216,7 +1255,7 @@ static readstat_error_t sav_parse_records_pass2(sav_ctx_t *ctx) {
                         goto cleanup;
                     }
                 }
-                if (io->read(data_buf, data_len, io->io_ctx) < data_len) {
+                if (data_len == 0 || io->read(data_buf, data_len, io->io_ctx) < data_len) {
                     retval = READSTAT_ERROR_PARSE;
                     goto cleanup;
                 }
@@ -1246,7 +1285,7 @@ static readstat_error_t sav_parse_records_pass2(sav_ctx_t *ctx) {
                             goto cleanup;
                         break;
                     case SAV_RECORD_SUBTYPE_LONG_VALUE_LABELS:
-                        retval = sav_parse_long_value_labels_record(data_buf, count, ctx);
+                        retval = sav_parse_long_value_labels_record(data_buf, size, count, ctx);
                         if (retval != READSTAT_OK)
                             goto cleanup;
                     default: /* misc. info */
@@ -1271,7 +1310,7 @@ static void sav_set_n_segments_and_var_count(sav_ctx_t *ctx) {
     int i;
     ctx->var_count = 0;
     for (i=0; i<ctx->var_index;) {
-        spss_varinfo_t *info = &ctx->varinfo[i];
+        spss_varinfo_t *info = ctx->varinfo[i];
         if (info->string_length) {
             info->n_segments = (info->string_length + 251) / 252;
         }
@@ -1281,22 +1320,22 @@ static void sav_set_n_segments_and_var_count(sav_ctx_t *ctx) {
     ctx->variables = readstat_calloc(ctx->var_count, sizeof(readstat_variable_t *));
 }
 
-static readstat_error_t sav_handle_variables(readstat_parser_t *parser, sav_ctx_t *ctx) {
+static readstat_error_t sav_handle_variables(sav_ctx_t *ctx) {
     int i;
     int index_after_skipping = 0;
     readstat_error_t retval = READSTAT_OK;
 
-    if (!parser->variable_handler)
+    if (!ctx->handle.variable)
         return retval;
 
     for (i=0; i<ctx->var_index;) {
         char label_name_buf[256];
-        spss_varinfo_t *info = &ctx->varinfo[i];
+        spss_varinfo_t *info = ctx->varinfo[i];
         ctx->variables[info->index] = spss_init_variable_for_info(info, index_after_skipping);
 
         snprintf(label_name_buf, sizeof(label_name_buf), SAV_LABEL_NAME_PREFIX "%d", info->labels_index);
 
-        int cb_retval = parser->variable_handler(info->index, ctx->variables[info->index],
+        int cb_retval = ctx->handle.variable(info->index, ctx->variables[info->index],
                 info->labels_index == -1 ? NULL : label_name_buf,
                 ctx->user_ctx);
 
@@ -1317,14 +1356,14 @@ cleanup:
     return retval;
 }
 
-static readstat_error_t sav_handle_fweight(readstat_parser_t *parser, sav_ctx_t *ctx) {
+static readstat_error_t sav_handle_fweight(sav_ctx_t *ctx) {
     readstat_error_t retval = READSTAT_OK;
     int i;
-    if (parser->fweight_handler && ctx->fweight_index >= 0) {
+    if (ctx->handle.fweight && ctx->fweight_index >= 0) {
         for (i=0; i<ctx->var_index;) {
-            spss_varinfo_t *info = &ctx->varinfo[i];
+            spss_varinfo_t *info = ctx->varinfo[i];
             if (info->offset == ctx->fweight_index - 1) {
-                if (parser->fweight_handler(ctx->variables[info->index], ctx->user_ctx) != READSTAT_HANDLER_OK) {
+                if (ctx->handle.fweight(ctx->variables[info->index], ctx->user_ctx) != READSTAT_HANDLER_OK) {
                     retval = READSTAT_ERROR_USER_ABORT;
                     goto cleanup;
                 }
@@ -1342,12 +1381,12 @@ readstat_error_t sav_parse_timestamp(sav_ctx_t *ctx, sav_file_header_record_t *h
     struct tm timestamp = { .tm_isdst = -1 };
 
     if ((retval = sav_parse_time(header->creation_time, sizeof(header->creation_time),
-                    &timestamp, ctx->error_handler, ctx->user_ctx)) 
+                    &timestamp, ctx->handle.error, ctx->user_ctx)) 
             != READSTAT_OK)
         goto cleanup;
 
     if ((retval = sav_parse_date(header->creation_date, sizeof(header->creation_date),
-                    &timestamp, ctx->error_handler, ctx->user_ctx)) 
+                    &timestamp, ctx->handle.error, ctx->user_ctx)) 
             != READSTAT_OK)
         goto cleanup;
 
@@ -1390,11 +1429,7 @@ readstat_error_t readstat_parse_sav(readstat_parser_t *parser, const char *path,
         goto cleanup;
     }
 
-    ctx->progress_handler = parser->progress_handler;
-    ctx->error_handler = parser->error_handler;
-    ctx->note_handler = parser->note_handler;
-    ctx->value_handler = parser->value_handler;
-    ctx->value_label_handler = parser->value_label_handler;
+    ctx->handle = parser->handlers;
     ctx->input_encoding = parser->input_encoding;
     ctx->output_encoding = parser->output_encoding;
     ctx->user_ctx = user_ctx;
@@ -1429,20 +1464,24 @@ readstat_error_t readstat_parse_sav(readstat_parser_t *parser, const char *path,
         goto cleanup;
     }
 
-    if (parser->info_handler) {
-        if (parser->info_handler(ctx->record_count == -1 ? -1 : ctx->row_limit,
-                    ctx->var_count, ctx->user_ctx) != READSTAT_HANDLER_OK) {
-            retval = READSTAT_ERROR_USER_ABORT;
-            goto cleanup;
-        }
-    }
-
-    if (parser->metadata_handler) {
+    if (ctx->handle.metadata) {
+        readstat_metadata_t metadata = {
+            .row_count = ctx->record_count == -1 ? -1 : ctx->row_limit,
+            .var_count = ctx->var_count,
+            .file_encoding = ctx->input_encoding,
+            .file_format_version = ctx->format_version,
+            .creation_time = ctx->timestamp,
+            .modified_time = ctx->timestamp,
+            .compression = ctx->compression,
+            .endianness = ctx->endianness
+        };
         if ((retval = readstat_convert(ctx->file_label, sizeof(ctx->file_label),
                         header.file_label, sizeof(header.file_label), ctx->converter)) != READSTAT_OK)
             goto cleanup;
 
-        if (parser->metadata_handler(ctx->file_label, ctx->input_encoding, ctx->timestamp, 2, ctx->user_ctx) != READSTAT_HANDLER_OK) {
+        metadata.file_label = ctx->file_label;
+
+        if (ctx->handle.metadata(&metadata, ctx->user_ctx) != READSTAT_HANDLER_OK) {
             retval = READSTAT_ERROR_USER_ABORT;
             goto cleanup;
         }
@@ -1450,13 +1489,13 @@ readstat_error_t readstat_parse_sav(readstat_parser_t *parser, const char *path,
 
     sav_parse_variable_display_parameter_record(ctx);
 
-    if ((retval = sav_handle_variables(parser, ctx)) != READSTAT_OK)
+    if ((retval = sav_handle_variables(ctx)) != READSTAT_OK)
         goto cleanup;
 
-    if ((retval = sav_handle_fweight(parser, ctx)) != READSTAT_OK)
+    if ((retval = sav_handle_fweight(ctx)) != READSTAT_OK)
         goto cleanup;
 
-    if (ctx->value_handler) {
+    if (ctx->handle.value) {
         retval = sav_read_data(ctx);
     }
     
