@@ -26,7 +26,8 @@ test_that("subsetting preserves attributes", {
   x <- labelled_spss(
     1:5, c("Good" = 1, "Bad" = 5),
     na_value = c(1, 2),
-    na_range = c(3, Inf)
+    na_range = c(3, Inf),
+    label = "Rating"
   )
   expect_identical(x, x[])
 })
